// ReedyTen - 10-channel reed transmitter (for modern R/C hardware) with oled display, encoder-navigated menus etc.
// Copyright (c) 2016, 2017 ceptimus.  Started August 2016
#include "eeprom.h"
#include "display.h"
#include "screens.h"
#include "controls.h"
#include "displayControls.h"
#include "encoder.h"
#include "beeper.h"
#include "cppm_out.h"
#include "timers.h"
#include "telemetry.h"

uint32_t smoothedA3; // when shifted 9 bits right, this is a smoothed version of Analog input 3 with 10 bit resolution (0-1023 range);
uint8_t batteryTenthsOfVolts; 
uint8_t batteryWarn = 0;

uint32_t prevMillis;

// signature for recognizing an intialized eeprom is 'Reed'
#define SIGNATURE 0x52656564UL

void setup(void) {
  uint8_t i;
  encoder::start();
  beeper::initialize();
  _display::start();
  bool initialized = true;
  uint32_t sig = eeprom::signature();
  if (sig != SIGNATURE) { // at first check eeprom appears uninitialized, but check it again and report what's found
    delay(1000);
    _display::displayStringSmall(0, 1, F("Checking EEPROM..."), 0);
    for (i = 0; i < 4; i++) {
      screens::hex(sig, 8, i / 2 + 1, (i % 2) * 9 + 2, 0);
      delay(1000);
      sig = eeprom::signature();
      if (sig == SIGNATURE) {
        break;    
      }
    }
    if (sig != SIGNATURE) {
      _display::displayStringSmall(3, 0, F("Press encoder to"), 0);
      _display::displayStringSmall(4, 0, F("initialize eeprom"), 0);
      _display::displayStringSmall(5, 0, F("(clears memories)"), 0);
      _display::displayStringSmall(6, 0, F("or move toggle switch"), 0);
      _display::displayStringSmall(7, 0, F("to bypass."), 0);
      PORTD |= 0xFE; // activate input pull ups on first 7 toggle inputs
      delay(100);
      while (true) {
        if (!digitalRead(A2)) { // encoder pressed
          initialized = false;
          break;  
        }
        if ((PIND & 0xFE) != 0xFE) { // either steering toggle, or  motor moved or up elevator (elevator toggle down)
          break; 
        }
      }
    }    
  }
  if (!initialized) {
    eeprom::gd.modelNum = 1;
    // TODO move builder-configurable values to #defines in config.h or similar
    eeprom::gd.batteryRange = 100;// theoretical value is 100 for a 10V input range (47K 47K divider into 5V input) - can be edited on Transmitter settings screen so no need to edit here
    eeprom::gd.lowBatteryAlarm = 71; // default alarm to 7.1 volts - can be edited on Transmitter settings screen so no need to edit here
    eeprom::putGD();
    for (i = 0; i < 13; i++) {
      eeprom::md.modelName[i] = ' ';
    }

    eeprom::md.timer[0] = 0x4000 | 300; // five minute Throttle > 25% Down
    eeprom::md.timer[1] = 0xA000 | 300; // Manual Up Silent (300 only to give default five minutes when user changes type - as for silent timers duration is ignored)

    for (i = 0; i < 4; i++) {
      eeprom::md.trimPos[i] = 0; // default all trims to zero (no trim on ele trim). Note 0 on trimPos[2] (throttle) is lowest trim - not centre
    }
    
    for (i = 0; i < 12; i++) {
      eeprom::md.endPoint[i] = 100; // default all end points to 100%.
    } 

    for (i = 0; i < 6; i++) {
      eeprom::md.subTrim[i] = 0;
    }

    eeprom::md.outputMix[0] = 0x41; // 1st output channel defaults to Bot(1), 2nd to Ele (4)
    eeprom::md.outputMix[1] = 0x23; // 3rd output channel defaults to Mot(3), 4th to Top (2)
    eeprom::md.outputMix[2] = 0x10; // 5th output channel defaults to Non(0), 6th to Bot (1)
     
    eeprom::md.reverse = B10000000; // all output channels default to normal (not reversed); mixer Steering Source defaults to top switch (for most fliers that's rudder for V-tail mixing)
    eeprom::md.eleTrimRange = 20; // 20% of total travel
    eeprom::md.trimRange = 20; // 20% of total travel

    eeprom::md.normServoTime = 150; // one point five seconds for end-to-end travel
    eeprom::md.progServoTime = 150; // three seconds for end-to-end travel
    
    eeprom::md.Aalarm = 0; // default to no telemetry (0) but 46 is a reasonable value for a 4-cell NiMH (4.6V) on A1    
    eeprom::md.Arange = 0; // 0 default to using A1 for model battery monitoring
    
    for (i = 2; i <= MAX_MODEL_NO; i++) {
      eeprom::putMD(i);  
    }

    eeprom::md.modelName[0] = 'M';
    eeprom::md.modelName[1] = 'o';
    eeprom::md.modelName[2] = 'd';
    eeprom::md.modelName[3] = 'e';
    eeprom::md.modelName[4] = 'l';
    eeprom::md.modelName[6] = '1';
    eeprom::putMD(1);
    eeprom::writeSignature(SIGNATURE);
  }

  eeprom::getGD();
  smoothedA3 = (uint32_t)analogRead(A3) << 9;
  batteryWarn = 0;
  eeprom::getMD(eeprom::gd.modelNum);
  controls::setServoSpeeds();
  controls::start();
  controls::scan(10); // initialize us for channels (especially throttle channel(s)) before starting cppm_out
  cppm_out::start();
  telemetry::start();
  beeper::lowModelVoltage = false;
  timers::initialize();

  _display::fillDisplay(' ' | 0x80);
  _display::displayStringSmall(0, 0, F("   Welcome to the    "), 0);
  _display::displayStringSmall(6, 14, F("V1.0.6"), 1); 
  _display::displayStringSmall(7, 0, F("  (c) 2017 ceptimus  "), 0); 
  _display::displayStringLarge(2, 2, F("ceptimus"), 1);
  _display::displayStringLarge(4, 1, F("Reedy Ten"), 1);
  delay(2000);
  prevMillis = millis();
}

uint8_t screen = 0;
uint8_t oldScreen = 0xFF;
uint8_t choice = 0;
bool editing = false;

void loop(void) {
  bool clicked;
  uint8_t i, mix;
  
  uint32_t now = millis();
  uint16_t ms = (uint16_t)(now - prevMillis);
  prevMillis = now;

  screen = beeper::loop(ms, screen);
  timers::loop(ms);
  controls::scan(ms);
  encoder::checkEncoder(ms);
  if (eeprom::md.Aalarm) { // if user has enabled telemetry for this model
    telemetry::check(ms);
    i = eeprom::md.Arange ? telemetry::a2 * eeprom::md.Arange / 255 : telemetry::a1 * 33 * 4 / 255; // model battery voltage in tenths of volts
    beeper::lowModelVoltage = i && (i <= eeprom::md.Aalarm); // sound alarm if telemetry being received and battery is low    
  }

  if (ms && ms < 512) { // read battery voltage and smooth taking into account loop time
    smoothedA3 = ((smoothedA3 * (512 - ms)) >> 9) + (uint32_t)analogRead(A3) * ms;
    batteryTenthsOfVolts = (uint8_t)(((smoothedA3 >> 9) * eeprom::gd.batteryRange) >> 10); 
    batteryWarn = batteryTenthsOfVolts < eeprom::gd.lowBatteryAlarm;
    if (batteryWarn) beeper::beep(ms); // sound buzzer continuously on low battery
  }

  if (screen != oldScreen) {
    if (eeprom::globalDataEdited) {
      eeprom::putGD();
    }
    if (eeprom::currentModelEdited) {
      eeprom::putMD(eeprom::gd.modelNum);
      beeper::lowModelVoltage = false;
    }
    if (screen == 0xFF) { // inactivity
      choice = 0;
      _display::fillDisplay(' ' | 0x80);
      _display::displayStringLarge(2, 2, F("Inactive"), 1);
      _display::displayStringLarge(4, 5, F("Alarm"), 1);
    } else {
      screens::fixedText(screen);
    }
    if ((screen != 1) && (screen != 2)) {
      screens::choice(screen, 1, &clicked);
    }
    switch (screen) {
      case 0: // Main Page
        displayControls::start();
        screens::displayCurrentModelName(4, 5);
        if (!eeprom::md.Aalarm) { // hide telemetry item labels for non-telemetry models
          screens::fTxt(45, 0121, 0);
          screens::fTxt(59, 0113, 0);
          screens::fTxt(59, 0173, 0);
        }
        if (!(eeprom::md.timer[0] & 0xE000)) { // hide Tmr1: label if inactive          
          screens::fTxt(41, 0065, 0);
        }
        if (!(eeprom::md.timer[1] & 0xE000)) { // hide Tmr2: label if inactive  TODO: hide S & R selections for inactive timer(s)         
          screens::fTxt(41, 0066, 0);
        }
        break;
      case 1: // Channels Monitor (%)
      case 2: // Channels Monitor (us)
        screens::prepareBargraphs();
        screens::ChLabel(2);
        break;
      case 5: // Copy Current Model
        screens::displayCurrentModelName(1, 2);
        screens::fTxt(78, 0072, 0);
        break;
      case 7: // Model Name
        screens::displayCurrentModelName(4, 2);
        break;
      case 8: // Timers
        timers::showType(0, 0022, 0);
        timers::showDuration(0, 3, 011, 0);
        timers::showType(1, 0025, 0);
        timers::showDuration(1, 6, 011, 0);
        break;
      case 9: // Servo Speeds
        screens::positiveDecimal(eeprom::md.normServoTime, 4, 2, 3, 016, 0);  
        screens::positiveDecimal(((uint16_t)eeprom::md.progServoTime) << 1, 4, 2, 6, 016, 0);  
        break;
      case 10: // Trim Ranges
        screens::positiveDecimal(eeprom::md.eleTrimRange, 2, 0, 3, 017, 0);  
        screens::positiveDecimal(eeprom::md.trimRange, 2, 0, 6, 017, 0);  
        break;
      case 11: // Mixer Steering Source
        screens::fTxt(eeprom::md.reverse & 0x80 ? 17 : 18, 0043, 0);
        break;      
      case 12: // Channel Mixing
        screens::ChLabel(1);
        for (i = 0; i < 6; i++) {
          mix = eeprom::md.outputMix[i >> 1];
          if (i & 0x01)
            mix >>= 4;
          mix &= 0x0F;
          screens::fTxt(mix + 60, 0031 + i, 0);
        }
        break;
      case 13: // Servo Reverse
        screens::ChLabel(2);
        for (i = 0; i < 6; i++) {
          screens::fTxt(eeprom::md.reverse & (0x01 << i) ? 77 : 76,  0051 + i, 0);
        }
        break;
      case 14: // End Points
        screens::ChLabel(8);
        for (i = 0; i < 12; i++) {
          screens::positiveDecimal(eeprom::md.endPoint[i], 3, 0, 1 + (i >> 1), i & 0x01 ? 020 : 013, 0);  
        }
        break;
      case 15: // Sub Trim
        screens::ChLabel(4);
        for (i = 0; i < 6; i++) {
          screens::signedDecimal(eeprom::md.subTrim[i], 3, 0, i + 1, 007, 0);  
        }
        break;
      case 16: // Telemetry
        screens::positiveDecimal(eeprom::md.Arange, 4, 1, 4, 017, 0);  
        screens::positiveDecimal(eeprom::md.Aalarm, 4, 1, 5, 017, 0);  
        break;
      case 17: // transmitter settings
        screens::positiveDecimal(eeprom::gd.batteryRange, 4, 1, 4, 017, 0);
        screens::positiveDecimal(eeprom::gd.lowBatteryAlarm, 4, 1, 5, 017, 0);
        break;
    }
    oldScreen = screen;
    editing = false;
  }
  if (screen == 0xFF) return; // if inactivity, exit loop()

  if ((screen == 0) || (screen == 17)) { // show live Transmitter battery voltage on Main Page and Transmitter Settings
    screens::positiveDecimal(batteryTenthsOfVolts, 4, 1, 2, 020, batteryWarn);
  } else if (screen == 16) { // telemetry
    i = telemetry::rssi <= 45;
    screens::positiveDecimal(telemetry::rssi, 3, 0, 2, 005, i); 
    screens::positiveDecimal(telemetry::a1, 3, 0, 1, 003, i); 
    screens::positiveDecimal(telemetry::a2, 3, 0, 1, 022, i);
    i = eeprom::md.Arange ? telemetry::a2 * eeprom::md.Arange / 255 : telemetry::a1 * 33 * 4 / 255; // model battery tenths of volts
    screens::positiveDecimal(i, 4, 1, 2, 020, i <= eeprom::md.Aalarm);
  }

  if (editing) { // turning encoder changes currently selected item
    int16_t change = encoder::change(0);

    if (change) {
      eeprom::currentModelEdited = true; // unless user is changing the transmitter settings - handled in that case 
      switch(screen) {
        case 7: // Model Name
          encoder::setMultiplier(5);
          change += eeprom::md.modelName[choice - 1];
          if (change < 32) {
            change = 32;
          } else if (change > 127) {
            change = 127;
          }
          eeprom::md.modelName[choice - 1] = (char)change;
          _display::displayCharSmall(4, 4 + choice, (char)change, 1);
          break;
        case 8: // Timers
          switch (choice) {
            case 1: // timer 1 type
            case 3: // timer 2 type
              encoder::setMultiplier(1);
              i = choice >> 1; // 0 for timer 1, 1 for timer 2
              change += (eeprom::md.timer[i] >> 13) & 0x0007;
              while (change < 0) change += 8;
              while (change > 7) change -= 8;
              eeprom::md.timer[i] = (eeprom::md.timer[i] & 0x1FFF) | ((uint16_t)change << 13);
              timers::reset(i);
              timers::showType(i, i ? 0025 : 0022, 1);
              break;
            case 2: // timer 1 duration
            case 4: // timer 2 duration
              encoder::setMultiplier(20);
              i = (choice - 1) >> 1; // 0 for timer 1, 1 for timer 2
              change += eeprom::md.timer[i] & 0x1FFF;
              if (change < 0) change = 0;
              else if (change > 5999) change = 5999;
              eeprom::md.timer[i] = (eeprom::md.timer[i] & 0xE000) | change;
              timers::reset(i);
              timers::showDuration(i, i ? 6 : 3, 011, 1);
              break;
          }
          break;
        case 9: // Servo Speeds
          switch (choice) {
            case 1: // normal servo travel time
              change += eeprom::md.normServoTime;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::md.normServoTime = (uint8_t)change;
              controls::setServoSpeeds();
              screens::positiveDecimal(change, 4, 2, 3, 016, 1);  
              break;
            case 2: // progressive servo travel time
              change += eeprom::md.progServoTime;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::md.progServoTime = (uint8_t)change;
              controls::setServoSpeeds();
              screens::positiveDecimal(change << 1, 4, 2, 6, 016, 1);              
              break;              
          }
          break;
        case 10: // Trim Ranges
          switch (choice) {
            case 1: // elevator trim range
              change += eeprom::md.eleTrimRange;
              if (change < 0) change = 0;
              else if (change > 50) change = 50;
              eeprom::md.eleTrimRange = (uint8_t)change;
              screens::positiveDecimal(change, 2, 0, 3, 017, 1);  
              break;
            case 2: // button trim range
              change += eeprom::md.trimRange;
              if (change < 0) change = 0;
              else if (change > 50) change = 50;
              eeprom::md.trimRange = (uint8_t)change;
              screens::positiveDecimal(change, 2, 0, 6, 017, 1);  
              break;
          }              
          break;
        case 11: // Mixer Steering Source
          if (change & 0x0001) {
            eeprom::md.reverse ^= 0x80;
            screens::fTxt(eeprom::md.reverse & 0x80 ? 17 : 18, 0043, 1);
          }
          break;
        case 12: // Channel Mixing
          mix = eeprom::md.outputMix[(choice - 1) >> 1];
          if (!(choice & 0x0001))
            mix >>= 4;
          mix &= 0x0F;
          change += mix;
          while (change < 0) change += 16;
          while (change > 15) change -= 16;
          eeprom::md.outputMix[(choice - 1) >> 1] = (eeprom::md.outputMix[(choice - 1) >> 1] & ((choice & 0x0001) ? 0xF0 : 0x0F)) | (choice & 0x01 ? change : change << 4);
          screens::fTxt(change + 60, 0030 + choice, 1);
          break;
        case 13: // Servo Reverse
          if (change & 0x0001) {
            eeprom::md.reverse ^= (0x01 << (choice - 1));
            screens::fTxt(eeprom::md.reverse & (0x01 << (choice - 1)) ? 77 : 76,  0050 + choice, 1);
          }
          break;
        case 14: // End Points
          change += eeprom::md.endPoint[choice - 1];
          if (change < -125) change = -125;
          if (change > 125) change = 125;
          eeprom::md.endPoint[choice - 1] = change;
          screens::positiveDecimal(change, 3, 0, 1 + ((choice - 1) >> 1), choice & 0x01 ? 013 : 020, 1);  
          break;
        case 15: // Sub Trim
          change += eeprom::md.subTrim[choice - 1];
          if (change < -50) change = -50;
          else if (change > 50) change = 50;
          eeprom::md.subTrim[choice - 1] = change;
          screens::signedDecimal(eeprom::md.subTrim[choice - 1], 3, 0, choice, 007, 1);
          break;
        case 16: // Telemetry
          switch (choice) {
            case 1: // Analogue Voltage Range
              change += eeprom::md.Arange;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::md.Arange = change;
              screens::positiveDecimal(change, 4, 1, 4, 017, 1);
              break;
            case 2: // Analogue Voltage Alarm
              change += eeprom::md.Aalarm;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::md.Aalarm = change;
              screens::positiveDecimal(change, 4, 1, 5, 017, 1);
              break;
          }
          break;
        case 17: // Transmitter Settings
          eeprom::currentModelEdited = false;
          eeprom::globalDataEdited = true;
          switch (choice) {
            case 1: // Analogue Voltage Range
              change += eeprom::gd.batteryRange;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::gd.batteryRange = change;
              screens::positiveDecimal(change, 4, 1, 4, 017, 1);
              break;
            case 2: // Analogue Voltage Alarm
              change += eeprom::gd.lowBatteryAlarm;
              if (change < 0) change = 0;
              else if (change > 255) change = 255;
              eeprom::gd.lowBatteryAlarm = change;
              screens::positiveDecimal(change, 4, 1, 5, 017, 1);
              break;
          }
          break;
      }
    }
    if (encoder::pressed) {
      encoder::setMultiplier(1); // no multiplier when navigating around screens
      editing = false;
    }
  } else { // moving cursor around posible selections
    choice = screens::choice(screen, choice, &clicked);
    switch (screen) {
      case 0: // Main Page      
        displayControls::refresh();
        timers::show(0, 5, 11, 0);
        timers::show(1, 6, 11, 0);
        if (eeprom::md.Aalarm) { // if user has enabled telemetry for this model          
          screens::positiveDecimal(telemetry::rssi, 3, 0, 1, 021, telemetry::rssi <= 45);
          i = eeprom::md.Arange ? telemetry::a2 * eeprom::md.Arange / 255 : telemetry::a1 * 33 * 4 / 255; // model battery tenths of volts
          screens::positiveDecimal(i, 4, 1, 3, 020, i <= eeprom::md.Aalarm);
        }
        if (clicked) {
          switch (choice) {
            case 0: // Main Menu
              screen = 3;
              break;
            case 1: // S timer 1
              timers::startStop(0);
              break;
            case 2: // R timer 1
              timers::reset(0);
              break;
            case 3: // S timer 2
              timers::startStop(1);
              break;
            case 4: // R timer 2
              timers::reset(1);
              break;
            case 5: // Mon
              choice = 0; screen = 1;
              break;
          }
        }
        break;
      case 1: // Channels Monitor (%)
        screens::displayChannels(0);
        if (clicked)
          screen = 2; // only one choice 0 on this screen
        break;
      case 2:
        screens::displayChannels(1);
        if (clicked) {
          choice = 5; screen = 0; // back to Main Page with 'Mon' choice selected
        }
        break;
      case 3: // Main Menu
        if (clicked) {
          switch (choice) {
            case 0: // Main Page
              screen = 0;
              break;
            case 1: // Select Model
              choice = eeprom::gd.modelNum - 1; screen = 4;
              break;
            case 2: // Model Settings
              choice = 0; screen = 6;
              break;
            case 3: // Copy Current Model
              choice = eeprom::gd.modelNum - 1; screen = 5;
              break;
            case 4: // Transmitter Settings
              choice = 0; screen = 17;
              break;
          }
        }
        break;
      case 4: // Select Model
        if (clicked) {
          eeprom::gd.modelNum = choice + 1;
          eeprom::getMD(choice + 1);
          beeper::lowModelVoltage = false;
          eeprom::globalDataEdited = true;
          controls::start(); // added for V1.0.1 to recentre dedicated reed elevator toggle switch when changing models.
          choice = screen = 0; 
        }
        break;
      case 5: // Copy Current Model
        if (clicked) {
          eeprom::putMD(choice + 1);
          eeprom::gd.modelNum = choice + 1;
          eeprom::getMD(choice + 1);
          eeprom::globalDataEdited = true;
          controls::start(); // added for V1.0.1 to recentre dedicated reed elevator toggle switch when changing models.
          choice = screen = 0; 
        }
        break;
      case 6: // Model Settings
        if (clicked) {
          switch (choice) {
            case 0: // Main menu
              choice = 2; screen = 3;
              break;
            default:
              screen += choice; choice = 0;
              break;
          }
        }
        break;
      case 17: // Transmitter Settings    
        if (clicked) {
          switch (choice) {
            case 0: // Main menu
              choice = 4; screen = 3;
              break;
            default:
              editing = true;
              encoder::setMultiplier(10);
              break;
          }
        }
        break;     
      default: // Model Settings Screens 7-16 
        if (clicked) {
          switch (choice) {
            case 0:
              choice = screen - 6; screen = 6;
              break;
            default:
              editing = true;
              encoder::setMultiplier(10); // set default multiplier for editing
              break;
          }
        }
        break;
    }
  }
}

