#include "controls.h"
#include "displayControls.h"
#include "eeprom.h"
#include "beeper.h"
#include "cppm_out.h"


int32_t controls::normServoSpeed, controls::progServoSpeed;
int32_t controls::posn[5];
int32_t controls::tPosn[4];
int32_t controls::output[6];
int16_t controls::trimTmr;
uint16_t controls::noTrimTmr;
int32_t controls::sweepPosn;
bool controls::sweepToggle;
bool controls::trimChanged;

void controls::start(void) {
#ifdef PPM_ON_D0
  pinMode(12, INPUT_PULLUP); // when outputting PPM on D0, then D12 becomes the trim button input, otherwise it's D0
#else
  pinMode(0, INPUT_PULLUP); // when outputting PPM on D0, then D12 becomes the trim button input, otherwise it's D0
#endif
  for (uint8_t i = 1; i <= 10; i++) {
    pinMode(i, INPUT_PULLUP); // the ten 'channel' inputs (both sides of five switches)
  }  
  for (uint8_t i = 0; i < 5; i++) {
    posn[i] = 0L; // centre all controls
  }
  tPosn[2] = posn[2] = 0x200000L; // except initialize motor control to maximum (position is in opposite sense to output because the motor down (idle) input comes before motor up on standard reed sets).
  sweepPosn = 0L;
  trimChanged = sweepToggle = false;
}

void controls::scan(uint16_t ms) { // ms is milliseconds elapsed since previous scan
  uint16_t pins = ((uint16_t)PINB) << 8 | PIND; // read all digital inputs into one word
  
  uint16_t mask = 0x0002; // mask for one switch contact
  bool oneChannelTrimmed = false; // prevent beeper doing double duty in the (rare) instance of  trimming two or more channels at once
  int32_t normDelta = normServoSpeed * ms;
  if (normDelta > 0x200000L)
    normDelta = 0x200000L;
  else if (normDelta < -0x200000L)
    normDelta = -0x200000L;
  int32_t progDelta = progServoSpeed * ms;
  if (progDelta > 0x200000L)
    progDelta = 0x200000L;
  else if (progDelta < -0x200000L)
    progDelta = -0x200000L;

  if (sweepToggle) {
    sweepPosn += normDelta;
    if (sweepPosn > 0x200000L) {
      sweepPosn = 0x400000L - sweepPosn;
      sweepToggle = false;
    }
  } else {
    sweepPosn -= normDelta;
    if (sweepPosn < -0x200000L) {
      sweepPosn =  -0x400000L - sweepPosn;
      sweepToggle = true;
    }
  }

#ifdef PPM_ON_D0
  if (trimChanged && ((pins & 0x1000) != 0)) { // if trim has been changed and trim button released
#else
  if (trimChanged && ((pins & 0x0001) != 0)) { // if trim has been changed and trim button released
#endif
    noTrimTmr += ms;
    if (noTrimTmr >= 10000) { // when no further trim changes for ten seconds
      trimChanged = false;
      eeprom::putMD(eeprom::gd.modelNum); // save model data to store trims
    }
  }
  
  for (uint8_t channel = 0; channel < 5; channel++) { // reed switches and trim processing
    bool a = pins & mask; // first contact of switch for channel
    mask <<= 1;
    bool b = pins & mask; // second contact of switch for channel
    mask <<= 1;
    int8_t stickPosn = (!a && b) ? 1 : (a && !b) ? -1 : 0; // stick position for channel -1, 0 or +1
    displayControls::stickPosn[channel] = stickPosn;
    if (stickPosn) beeper::activity(); // reset inactivity alarm if stick moved (or held) away from centre position
#ifdef PPM_ON_D0
    if ((pins & 0x1000) == 0) { // if trim button pressed alter trim rather than posn
#else
    if ((pins & 0x0001) == 0) { // if trim button pressed alter trim rather than posn
#endif
      beeper::activity(); // trim button pressed or held also resets inactivity alarm
      if ((pins & 0x01FE) == 0x01FE) { // none of the four main switches are activated - reset trim timer
        trimTmr = FIRST_TRIM_ms;
      } else if (stickPosn && channel < 5) { // at least this one of the four main switches is active.
        trimChanged = true;
        noTrimTmr = 0;
        int8_t minTrim, midTrim, maxTrim;
        if (channel == 2) {
          minTrim = -127;
          midTrim = -64;
          maxTrim = 0;
        } else {
          minTrim = -63;
          midTrim = 0;
          maxTrim = 63;
        }
        if ((trimTmr == FIRST_TRIM_ms) || (ms >= (uint16_t)trimTmr)) {
          eeprom::md.trimPos[channel] += stickPosn;
          if (!oneChannelTrimmed) beeper::beep(TRIM_BEEP_ms);
          
          if (eeprom::md.trimPos[channel] == midTrim) { // longer beep and auto-repeat delay at center
            beeper::beep(5 * TRIM_BEEP_ms);
            trimTmr = FIRST_TRIM_ms;
          }
          else if (eeprom::md.trimPos[channel] > maxTrim) { 
            eeprom::md.trimPos[channel] = maxTrim;
            if (!oneChannelTrimmed) beeper::beep(AUTO_REPEAT_TRIM_ms - TRIM_BEEP_ms); // limit reached - keep beeper active till operator lets go of switch or button
          }
          else if (eeprom::md.trimPos[channel] < minTrim) { 
            eeprom::md.trimPos[channel] = minTrim;
            if (!oneChannelTrimmed) beeper::beep(AUTO_REPEAT_TRIM_ms - TRIM_BEEP_ms);
          }
          displayControls::trimPosn[channel] = (eeprom::md.trimPos[channel] - midTrim) / 7;
        }
        if (!oneChannelTrimmed) {
          trimTmr -= ms;
          oneChannelTrimmed = true;
        }
        if (trimTmr <= 0) trimTmr += AUTO_REPEAT_TRIM_ms;
        stickPosn = 0; // when trim button pressed, other than trimming action, controls behave as though the stick is centred.  Special reed elevator trim channel is not in this loop, so still works normally when trim button pressed 
      }       
    } else { // trim button not pressed - reset trim timer
      trimTmr = FIRST_TRIM_ms;
    }
    int32_t servoSpeed = (channel == 2) || (channel == 4) ? progDelta : normDelta;

    bool stopAtCentre = false;
    if (!stickPosn) { // switch released
      if ((channel != 2) && (channel != 4)) {
        servoSpeed *= -sgn(posn[channel]); // return to centre when switch released for non-progressive channels
        stopAtCentre = true;
      } else {
        servoSpeed = 0L; // stay at current positon when switch released on progressive channels
      }
    } else { // switch activated
      servoSpeed *= (int32_t)stickPosn;
    }
    posn[channel] += servoSpeed;
    if      (posn[channel] >  0x200000L) posn[channel] =  0x200000L;
    else if (posn[channel] < -0x200000L) posn[channel] = -0x200000L;
    else if (stopAtCentre && (sgn(posn[channel]) == sgn(servoSpeed))) posn[channel] = 0L; // stop at centre when auto-returning to centre
#ifdef A6_THROTTLE_KILL
    bool throttleKilled = false;
    if ((channel == 2) && (analogRead(A6) < 512)) {
      throttleKilled = true;
      posn[2] = 0x200000L;
    }
#endif
    displayControls::servoPosn[channel] = (int8_t)((posn[channel] + 0x20000L) >> 18);
    if (channel < 4) { // apply normal eeprom trims to channels
      int32_t effect = 333L; // 50% range with full 63 trim steps should give offset of 2^20.  2^20 / (50 * 63) = 332.88
      if (channel == 2) effect = (effect * ((posn[channel] >> 8) + 0x2000L)) >> 14; // throttle trim only applies at bottom end and washes out to nothing at full throttle 
      tPosn[channel] = posn[channel] + effect * eeprom::md.trimRange * eeprom::md.trimPos[channel]; 
      if (channel == 3) tPosn[channel] += posn[4] * eeprom::md.eleTrimRange / 100; // apply special reed trim channel to elevator channel
      if      (tPosn[channel] >  0x200000L) tPosn[channel] =  0x200000L;
      else if (tPosn[channel] < -0x200000L) tPosn[channel] = -0x200000L;
#ifdef A6_THROTTLE_KILL
      if (throttleKilled) tPosn[2] = 0x200000L; // throttle kill switch overrides throttle trim setting (behaves as if throttle trim is set to minimum)
#endif
    }
  } // reed switches and trim processing
  
  for (uint8_t channel = 0; channel < 6; channel++) { // output mixing, end-points, servo reverse and sub-trims
    uint8_t mix = eeprom::md.outputMix[channel >> 1];
    if (channel & 0x01)
      mix >>= 4;
      
    mix &= 0x0F;
    int32_t steer = eeprom::md.reverse & 0x80 ? tPosn[1] : tPosn[0]; // choose either top switch or bottom switch for mixer steering source
    switch (mix) { // note signs of throttle and elevator inputs are reversed to correct reed order to modern standard output directions
      case 0: // unused channel
        output[channel] = 0L;
        break;
      case 1: // bottom steering switch
        output[channel] = tPosn[0];
        break;
      case 2: // top steering switch
        output[channel] = tPosn[1];
        break;
      case 3: // throttle
        output[channel] = -tPosn[2];
        break;
      case 4: // elevator
        output[channel] = -tPosn[3];
        break;
      case 5: // 50% steer + 50% elevator
        output[channel] = (steer >> 1) - (tPosn[3] >> 1);
        break;
      case 6: // 50% steer - 50% elevator
        output[channel] = (steer >> 1) + (tPosn[3] >> 1);
        break;
      case 7: // 60% steer + 40% elevator
        output[channel] = (steer * 6 / 10) - (tPosn[3] * 4 / 10);
        break;
      case 8: // 60% steer - 40% elevator
        output[channel] = (steer * 6 / 10) + (tPosn[3] * 4 / 10);
        break;
      case 9: // 40% steer + 60% elevator
        output[channel] = (steer * 4 / 10) - (tPosn[3] * 6 / 10);
        break;
      case 10: // 40% steer - 60% elevator
        output[channel] = (steer * 4 / 10) + (tPosn[3] * 6 / 10);
        break;
      case 11: // 70% steer + 30% elevator
        output[channel] = (steer * 7 / 10) - (tPosn[3] * 3 / 10);
        break;
      case 12: // 70% steer - 30% elevator
        output[channel] = (steer * 7 / 10) + (tPosn[3] * 3 / 10);
        break;
      case 13: // 30% steer + 70% elevator
        output[channel] = (steer * 3 / 10) - (tPosn[3] * 7 / 10);
        break;
      case 14: // 30% steer - 70% elevator
        output[channel] = (steer * 3 / 10) + (tPosn[3] * 7 / 10);
        break;
      case 15: // sweep
        output[channel] = sweepPosn;
        break;        
    }
    uint8_t downLeft = eeprom::md.endPoint[(channel << 1)];
    uint8_t upRight = eeprom::md.endPoint[(channel << 1) + 1];

    
    output[channel] *= output[channel] >= 0L ? upRight : downLeft;
    output[channel] /= 100L;

    int32_t downLeftLimit, upRightLimit;
    
    if (eeprom::md.reverse & (0x01 << channel)) {
      output[channel] = -output[channel];
      upRightLimit = (0x200000L * downLeft) / 100L;
      downLeftLimit = (-0x200000L * upRight) / 100L;
    } else {
      upRightLimit = (0x200000L * upRight) / 100L;
      downLeftLimit = (-0x200000L * downLeft) / 100L;
    }
    
    output[channel] += eeprom::md.subTrim[channel] * 20972L; // each 1% of subTrim is 0x200000 / 100 units = 20971.52
    
    if (output[channel] < downLeftLimit) {
      output[channel] = downLeftLimit;
    } else if (output[channel] > upRightLimit) {
      output[channel] = upRightLimit;
    }
    
    cppm_out::set(channel, output[channel] >> 11);
  }
}

bool controls::throttleUp(void) { // returns true if throttle set >= 10% of full range (>-80% on monitor)
  return tPosn[2] <= 0x199999L;
}

void controls::setServoSpeeds(void) {
  normServoSpeed = !eeprom::md.normServoTime ? 0x80000L : 0x400000L / (((int32_t)eeprom::md.normServoTime) * 10L);
  progServoSpeed = !eeprom::md.progServoTime ? 0x40000L : 0x400000L / (((int32_t)eeprom::md.progServoTime) * 20L);
}

