#include "cppm_out.h"

// pulse spacings in half-microseconds, all have 600 subtracted to allow for 300 us pulse
volatile uint16_t cppm_out::pulseWidth[6]; 

void cppm_out::start(void) { // start driving six-channel cppm on D12 (or optionally on D0 if #define PPM_ON_D0 is enabled in config.h)

#ifdef PPM_ON_D0
// with OPEN_OUTPUT there's no need for an RF-module-protecting Schottky diode
# ifdef OPEN_OUTPUT 
  PORTD &= ~_BV(0); // preseting D0 to go low only when it's enabled as an output
#  ifdef PULSES_LOW
  DDRD &= ~_BV(0); // for low-going pulses, normal state of pin is high impedance (input) and the RF module will pull it high
#  else
  DDRD |= _BV(0); // for high-going pulses, normal state of pin is output - and as the PORTD bit is permanently low this will be a low output
#  endif
# else
// OPEN_OUTPUT is not defined so we set up the pin as a regular active output
  pinMode(0, OUTPUT);
#  ifdef PULSES_LOW
  digitalWrite(0, HIGH); // for low-going pulses, normal state of pin is high
#  else
  digitalWrite(0, LOW); // for high-going pulses, normal state of pin is low
#  endif
# endif
#else
// with OPEN_OUTPUT there's no need for an RF-module-protecting Schottky diode
# ifdef OPEN_OUTPUT 
  PORTB &= ~_BV(4); // preseting D12 to go low only when it's enabled as an output
#  ifdef PULSES_LOW
  DDRB &= ~_BV(4); // for low-going pulses, normal state of pin is high impedance (input) and the RF module will pull it high
#  else
  DDRB |= _BV(4); // for high-going pulses, normal state of pin is output - and as the PORTB bit is permanently low this will be a low output
#  endif
# else
// OPEN_OUTPUT is not defined so we set up the pin as a regular active output
  pinMode(12, OUTPUT);
#  ifdef PULSES_LOW
  digitalWrite(12, HIGH); // for low-going pulses, normal state of pin is high
#  else
  digitalWrite(12, LOW); // for high-going pulses, normal state of pin is low
#  endif
# endif
#endif
    
  // initialize counter 2 for pulse output duties
  TCCR2A = 0x02; // CTC mode (Clear Timer on Compare Match)
  TCCR2B = 0x02; // prescale = clk/8: count at 2MHz
  OCR2A = 0xFF; // default count target (256 counts per cycle)
  TIMSK2 = 0x02; // Compare Match A Interrupt Enable
}

// driving 6-channel cppm on Digital 12 (PB4) using Timer 2 interrupts - ceptimus 2015-06-25
// pulses are all 300us wide. From one +ve edge to next separated by 800us to 2200us.  Frame is 20ms long
ISR(TIMER2_COMPA_vect) {
  static uint8_t cycles = 0;
  static uint8_t finalCount;
  static uint8_t stage = 0;
  static uint16_t syncWidth = 35800; // whole frame is 20,000 us.  subtract 7 x 300 us for the 7 pulses in a 6-channel frame
  if (!cycles) { // time to switch an output
    uint16_t width;
    if (!(stage & 0x01)) { // at even stages, start 300 us pulse
#ifdef PPM_ON_D0
# ifdef OPEN_OUTPUT
#  ifdef PULSES_LOW
      DDRD |= _BV(0); // enable D0 as output.  Port D bit 0 was initialized to low, so this makes D0 pull low - pulse train is mostly high with low pulses
#  else
      DDRD &= ~_BV(0); // disable D0 as output thus allowing the RF module to pull it high
#  endif
# else
#  ifdef PULSES_LOW
      PORTD &= ~_BV(0); // active output with low-going pulses - drive the output low
#  else
      PORTD |= _BV(0); // active output with high-going pulses - drive the output high
#  endif
# endif
#else
# ifdef OPEN_OUTPUT
#  ifdef PULSES_LOW
      DDRB |= _BV(4); // enable D12 as output.  Port B bit 4 was initialized to low, so this makes D12 pull low - pulse train is mostly high with low pulses
#  else
      DDRB &= ~_BV(4); // disable D12 as output thus allowing the RF module to pull it high
#  endif
# else
#  ifdef PULSES_LOW
      PORTB &= ~_BV(4); // active output with low-going pulses - drive the output low
#  else
      PORTB |= _BV(4); // active output with high-going pulses - drive the output high
#  endif
# endif
#endif
      width = 600; // pulse time in half-micro seconds
    }
    else { // at odd stages end 300 us pulse
#ifdef PPM_ON_D0
# ifdef OPEN_OUTPUT
#  ifdef PULSES_LOW
      DDRD &= ~_BV(0); // disable D0 as output thus allowing the RF module to pull it high
#  else
      DDRD |= _BV(0); // enable D0 as output.  Port D bit 0 was initialized to low, so this makes D0 pull low - pulse train is mostly low with high pulses
#  endif
# else
#  ifdef PULSES_LOW
      PORTD |= _BV(0); // active output with high-going pulses - drive the output high
#  else
      PORTD &= ~_BV(0); // active output with low-going pulses - drive the output low
#  endif
# endif
#else
# ifdef OPEN_OUTPUT
#  ifdef PULSES_LOW
      DDRB &= ~_BV(4); // disable D12 as output thus allowing the RF module to pull it high
#  else
      DDRB |= _BV(4); // enable D12 as output.  Port B bit 5 was initialized to low, so this makes D12 pull low - pulse train is mostly low with high pulses
#  endif
# else
#  ifdef PULSES_LOW
      PORTB |= _BV(4); // active output with high-going pulses - drive the output high
#  else
      PORTB &= ~_BV(4); // active output with low-going pulses - drive the output low
#  endif
# endif
#endif
      if (stage == 13) { // final stage so now output the sync pulse
        width = syncWidth;
        syncWidth = 35800;  // 50Hz frame is 40,000 counts wide, but subtract the 7 x 300 us pulses
      }
      else {
        width = cppm_out::pulseWidth[stage / 2];
        syncWidth -= width;
      }
    }
    if ((width & 0x00FF) < 50) { // if finalCount would be less than 50 with the default target at 0xFF
      uint8_t adjust = (50 - (width & 0x00FF)) / (width >> 8) + 1;
      OCR2A = 0x100 - adjust;
      cycles = width >> 8;
      finalCount = (width & 0xFF) + adjust * cycles - 1;
    }
    else { // okay to use default 0xFF count value for all but last cycle
      OCR2A = 0xFF;
      cycles = width >> 8;
      finalCount = (width & 0xFF) - 1;
    }  
    if (++stage > 13) { // move to next stage (14 stages for the 14 pulse edges)
        stage = 0;
    }
  }
  else if (cycles-- == 1) { // last cycle, load the remaining count target
    OCR2A = finalCount;
  }
}

void cppm_out::set(uint8_t n, int16_t p) {
  p += 2400; // nominal ppm (like edge to like edge) width is 1500 us, but that includes the 300 us  pulse
  if (p < 1000) { // limit range to 800us - 2200us ( x 2 - 600 to convert to counts less 600 for pulse)
    p = 1000;
  }
  else if (p > 3800) {
    p = 3800;
  }
  volatile uint16_t *ptr;
  ptr = pulseWidth + n;
  uint8_t SaveSREG = SREG;
  cli();
  *ptr = p;
  SREG = SaveSREG; // this is more complicated than just cli() and sei() but allows this function to be called from main program OR from within another interrupt
}

void cppm_out::set_us(uint8_t n, int16_t us) {
  set(n, (us - 1500) * 2);
}

int16_t cppm_out::get_us(uint8_t n) {
  int16_t p = (int16_t)pulseWidth[n] - 2400;
  return p / 2 + 1500;
}


