#include "encoder.h"
#include "beeper.h"

uint32_t encoder::lastTurnMillis;
uint8_t encoder::state;
int16_t encoder::count;
int16_t encoder::debounce;
uint16_t encoder::multiplier;
boolean encoder::pressed;


ISR(PCINT1_vect) { // pinchange occured on A0 or A1
  // least significant 4 bits in state are oldA1,oldA0,A1,A0
  encoder::state = ((encoder::state & 0x03) << 2) | (PINC & 0x03);
  switch(encoder::state) {
    case B0001: // +ve transition A0 while A1 low
      encoder::count--;
      break;
    case B0010: // +ve transition on A1 while A0 low
      encoder::count++;
      break;
    case B0100: // -ve transition on A0 while A1 low
      encoder::count++;
      break;
    case B0111: // +ve transition A1 while A0 high
      encoder::count--;
      break;
    case B1000: // -ve transition A1 while A0 low
      encoder::count--;
      break;
    case B1011: // +ve transition on A0 while A1 high
      encoder::count++;
      break;
    case B1101: // -ve transition on A1 while A0 high
      encoder::count++;
      break;
    case B1110: // -ve transition on A0 while A1 high
      encoder::count--;
      break;
  }
}

void encoder::start(void) {
    pinMode(A0, INPUT_PULLUP);
    pinMode(A1, INPUT_PULLUP);
    pinMode(A2, INPUT_PULLUP);
    state = PINC & 0x03; // initialize state ready for first pinchange interrupt
    count = 0;
    multiplier = 1;
    pressed = false;
    debounce = 0;
    PCMSK1 = 0x03; // mask bits of PORTC (A0 - A7) so that only A0 and A1 generate pinchange interrupts
    PCICR |= 0x02; // enable pinchange interrupts for PORTC
    lastTurnMillis = millis();
}

void encoder::checkEncoder(uint16_t ms) { // ms is milliseconds since previous call
  static boolean oldButton = false; // true if previously pressed
  boolean button; // true when pressed (debounced)

  if (!(PINC & 0x04)) { // undebounced press
    button = true; // comes on at first contact - debouncing is on release
    debounce = 100;
  } else if (debounce > 0) {
    button = true;
    debounce -= ms;
  } else {
    button = false;
  }
  if ((pressed = button && !oldButton)) { // pressed is true for one scan on first -ve edge of A2 after A2 has been high for 100mS
    cli(); count = 0; sei(); // set count to zero to remove any hysteresis deadband.
    beeper::activity(); // reset inactivity alarm when encoder pressed
  }
  oldButton = button;
}

void encoder::setMultiplier(uint16_t ratio) {
  if (multiplier != ratio) {
    multiplier = ratio;
  }
}

int16_t encoder::change(int16_t value) {
  int16_t delta;
  cli(); delta = count; sei();
  delta /= ENCODER_COUNTS_PER_CLICK;
  if (delta) {
    uint32_t now = millis();
    int32_t elapsed = now - lastTurnMillis;
    lastTurnMillis = now;
    if (elapsed > 200L)
      value += delta;
    else if (elapsed > 100L)
      value += min(2, multiplier) * delta;
    else if (elapsed > 50L)
      value += min(5, multiplier) * delta;
    else if (elapsed > 25L)
      value += min(10, multiplier) * delta;
    else
      value += multiplier * delta;
      
    delta *= ENCODER_COUNTS_PER_CLICK;
    cli(); count -= delta; sei();
  }
  if (delta) beeper::activity(); // reset inactivity alarm when encoder turned
  return value;
}

